@@ -19,7 +19,7 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 *********************************************************************/
 #include "datadevice_interface.h"
 #include "datadevicemanager_interface.h"
-#include "dataoffer_interface.h"
+#include "dataoffer_interface_p.h"
 #include "datasource_interface.h"
 #include "display.h"
 #include "resource_p.h"
@@ -48,11 +48,15 @@ public:
     SurfaceInterface *icon = nullptr;
 
     DataSourceInterface *selection = nullptr;
+    QMetaObject::Connection selectionUnboundConnection;
+    QMetaObject::Connection selectionDestroyedConnection;
 
     struct Drag {
         SurfaceInterface *surface = nullptr;
         QMetaObject::Connection destroyConnection;
         QMetaObject::Connection pointerPosConnection;
+        QMetaObject::Connection sourceActionConnection;
+        QMetaObject::Connection targetActionConnection;
         quint32 serial = 0;
     };
     Drag drag;
@@ -101,11 +105,14 @@ void DataDeviceInterface::Private::startDrag(DataSourceInterface *dataSource, Su
         return;
     }
     // TODO: source is allowed to be null, handled client internally!
+    Q_Q(DataDeviceInterface);
     source = dataSource;
+    if (dataSource) {
+        QObject::connect(dataSource, &Resource::aboutToBeUnbound, q, [this] { source = nullptr; });
+    }
     surface = origin;
     icon = i;
     drag.serial = serial;
-    Q_Q(DataDeviceInterface);
     emit q->dragStarted();
 }
 
@@ -119,11 +126,27 @@ void DataDeviceInterface::Private::setSelectionCallback(wl_client *client, wl_re
 
 void DataDeviceInterface::Private::setSelection(DataSourceInterface *dataSource)
 {
+    if (dataSource && dataSource->supportedDragAndDropActions() && wl_resource_get_version(dataSource->resource()) >= WL_DATA_SOURCE_ACTION_SINCE_VERSION) {
+        wl_resource_post_error(dataSource->resource(), WL_DATA_SOURCE_ERROR_INVALID_SOURCE, "Data source is for drag and drop");
+        return;
+    }
     Q_Q(DataDeviceInterface);
+    QObject::disconnect(selectionUnboundConnection);
+    QObject::disconnect(selectionDestroyedConnection);
+    if (selection) {
+        selection->cancel();
+    }
     selection = dataSource;
     if (selection) {
+        auto clearSelection = [this] {
+            setSelection(nullptr);
+        };
+        selectionUnboundConnection = QObject::connect(selection, &Resource::unbound, q, clearSelection);
+        selectionDestroyedConnection = QObject::connect(selection, &QObject::destroyed, q, clearSelection);
         emit q->selectionChanged(selection);
     } else {
+        selectionUnboundConnection = QMetaObject::Connection();
+        selectionDestroyedConnection = QMetaObject::Connection();
         emit q->selectionCleared();
     }
 }
@@ -187,7 +210,12 @@ DataSourceInterface *DataDeviceInterface::selection() const
 void DataDeviceInterface::sendSelection(DataDeviceInterface *other)
 {
     Q_D();
-    auto r = d->createDataOffer(other->selection());
+    auto otherSelection = other->selection();
+    if (!otherSelection) {
+        sendClearSelection();
+        return;
+    }
+    auto r = d->createDataOffer(otherSelection);
     if (!r) {
         return;
     }
@@ -213,6 +241,13 @@ void DataDeviceInterface::drop()
         return;
     }
     wl_data_device_send_drop(d->resource);
+    if (d->drag.pointerPosConnection) {
+        disconnect(d->drag.pointerPosConnection);
+        d->drag.pointerPosConnection = QMetaObject::Connection();
+    }
+    disconnect(d->drag.destroyConnection);
+    d->drag.destroyConnection = QMetaObject::Connection();
+    d->drag.surface = nullptr;
     client()->flush();
 }
 
@@ -230,12 +265,24 @@ void DataDeviceInterface::updateDragTarget(SurfaceInterface *surface, quint32 se
         disconnect(d->drag.destroyConnection);
         d->drag.destroyConnection = QMetaObject::Connection();
         d->drag.surface = nullptr;
+        if (d->drag.sourceActionConnection) {
+            disconnect(d->drag.sourceActionConnection);
+            d->drag.sourceActionConnection = QMetaObject::Connection();
+        }
+        if (d->drag.targetActionConnection) {
+            disconnect(d->drag.targetActionConnection);
+            d->drag.targetActionConnection = QMetaObject::Connection();
+        }
         // don't update serial, we need it
     }
     if (!surface) {
+        if (auto s = d->seat->dragSource()->dragSource()) {
+            s->dndAction(DataDeviceManagerInterface::DnDAction::None);
+        }
         return;
     }
-    DataOfferInterface *offer = d->createDataOffer(d->seat->dragSource()->dragSource());
+    auto *source = d->seat->dragSource()->dragSource();
+    DataOfferInterface *offer = d->createDataOffer(source);
     d->drag.surface = surface;
     if (d->seat->isDragPointer()) {
         d->drag.pointerPosConnection = connect(d->seat, &SeatInterface::pointerPosChanged, this,
@@ -266,6 +313,30 @@ void DataDeviceInterface::updateDragTarget(SurfaceInterface *surface, quint32 se
     const QPointF pos = d->seat->dragSurfaceTransformation().map(d->seat->pointerPos());
     wl_data_device_send_enter(d->resource, serial, surface->resource(),
                               wl_fixed_from_double(pos.x()), wl_fixed_from_double(pos.y()), offer ? offer->resource() : nullptr);
+    if (offer) {
+        offer->d_func()->sendSourceActions();
+        auto matchOffers = [source, offer] {
+            DataDeviceManagerInterface::DnDAction action{DataDeviceManagerInterface::DnDAction::None};
+            if (source->supportedDragAndDropActions().testFlag(offer->preferredDragAndDropAction())) {
+                action = offer->preferredDragAndDropAction();
+            } else {
+                if (source->supportedDragAndDropActions().testFlag(DataDeviceManagerInterface::DnDAction::Copy) &&
+                    offer->supportedDragAndDropActions().testFlag(DataDeviceManagerInterface::DnDAction::Copy)) {
+                    action = DataDeviceManagerInterface::DnDAction::Copy;
+                } else if (source->supportedDragAndDropActions().testFlag(DataDeviceManagerInterface::DnDAction::Move) &&
+                    offer->supportedDragAndDropActions().testFlag(DataDeviceManagerInterface::DnDAction::Move)) {
+                    action = DataDeviceManagerInterface::DnDAction::Move;
+                } else if (source->supportedDragAndDropActions().testFlag(DataDeviceManagerInterface::DnDAction::Ask) &&
+                    offer->supportedDragAndDropActions().testFlag(DataDeviceManagerInterface::DnDAction::Ask)) {
+                    action = DataDeviceManagerInterface::DnDAction::Ask;
+                }
+            }
+            offer->dndAction(action);
+            source->dndAction(action);
+        };
+        d->drag.targetActionConnection = connect(offer, &DataOfferInterface::dragAndDropActionsChanged, offer, matchOffers);
+        d->drag.sourceActionConnection = connect(source, &DataSourceInterface::supportedDragAndDropActionsChanged, source, matchOffers);
+    }
     d->client->flush();
 }
 
